#include <stdio.h>	// Copyright by Mikhail Ranish, 2003 (www.ranish.com)
#include <stdlib.h>	// Realeased under [ Attribution-ShareAlike License ]
#include <string.h>	//     http://creativecommons.org/licenses/by-sa/1.0/

// Blocksort implementation of the first step of BWT compression algorithm.
// It has 8*n memory efficiency and execution time of n+sum m*log(m) where m
// is the match lenght for every pair of the matching strings.
//
// Since m is usually limited by data type and not the block size n
// the algorithm is practicly linear in time in respect to the block size n
// Still the worst case is n*log(n) for the file of a repeated character.
//
// It is a modification of radix blocksort described by Larsson and Sadakane
// in "Faster Suffix Sorting" and essentially replaces Ternary-split Quicksort
// for an unsorted group with a linear-time sorting procedure using linked lists.
// Instead of implementing quicksort for the unsorted portions of the index it does
// it in a linear time by building a linked list of an unsorted group in the sorted order.
// Once a portion of the index becomes sorted it is not processed any more.
//
// In addition to the blocksort this file contains a variation of
// Distance Coding and algorithms to undo both DC and BWT.
//
// Run  block.exe <filename>   or   block.exe -s Mary had a little lamb... :)
//

#define F   (256)
#define N   (1L<<16)
#define R   (N-1)
#define NN  (N+N)
#define NNN (N+N+N)

unsigned char b[N];

int t[N+F+2], s[N+F+2];

void main(int argc, char **argv)
{
 int c, m, n;
 unsigned long C=0;

 FILE *fp;
 if( argc>=3 && strcmp(argv[1],"-s")==0 )
   {
    b[0]=0;for(c=2;c<argc;c++){strcat(b,argv[c]);strcat(b," ");}
    n=strlen(b)-1;
   }
 else{   
 if( argc!=2 || (fp=fopen(argv[1],"rb"))==0 || (n=fread(b,1,N,fp))==-1 )
   { fprintf(stderr,"Error opening or reading input file\n"); exit(1); }
 fclose(fp);
 }

//void blocksort(char *b, int *s, int *t, int n) // b[] could overly s[]
{
 int i, j, si, sj, g, k, c, d, m, f[F+3], q[F+3];
 
 for( i=0 ; i<F+2 ; i++ ) f[i]=0;
 for( i=0 ; i<n   ; i++ ) f[b[i]+1]++;
 for( i=0 ; i<F   ; i++ ) f[i+1]+=f[i];
 for( i=0 ; i<n-1 ; i++ ) t[i]=f[b[i]]; t[n-1]=(f[b[n-1]+1]-1)|NNN;
 for( i=0 ; i<n   ; i++ ) s[i]=i;

 // s[]  pointers to the strings sorted to the best knowledge (up to k)
 // t[]  sorted position (in the array s[]) of the string

 // t[i]&R==t[j]&R  if b[i],b[j] are unsorted in respect to each other
 // t[i]&NNN        if t[i] is uniq == finalised sorted position of b[i]

 // s[i]&NNN        when s[i] no longer index into t[], but index into
 //		    s[] to the last s[] that corresponds to the sorted group

 for( k=1 ; k<n && s[0]<NNN+n-1 ; k+=k )
    {
     for( i=0 ; i<n ; i++, C++ ) if( s[i]>=NNN ) { i=s[i]&R; }
       else { 
              while( (sj=s[j=t[si=s[i]&R]&R]&R)>=0 && (t[sj]&R)!=j ||
                      t[si]<NNN && (t[sj+k]&R)>(t[si+k]&R) )
                 { s[i]=sj; s[j]=si|NN; C++; }
              s[i]=s[i]&R;
              s[j]=s[j]&R|NN;
             }
      
     for( i=0 ; i<n ; i++, C++ ) if( s[i]>=NNN ) { i=s[i]&R; }
       else if( t[si=s[i]&R]>=NNN && (si<k || t[si-k]>=NNN) )
         {
          j=(i>0   && s[i-1]>=NNN ? s[i-1]&R : i);
          i=(i<n-1 && s[i+1]>=NNN ? s[i+1]&R : i);
          s[i]=j|NNN;
          s[j]=i|NNN;
         }
       else if( s[i]>=NN ) t[s[i]&R]|=NN;
       else while( s[i]<N )
               { j=++t[s[t[si=s[i]&R]&R]&R]&R; s[i]=s[j]; s[j]=si|N; C++; }

     for( g=i=n ; i-->0 ; C++ ) if( s[i]>=NNN  ) { g=i=s[i]&R; }
     else
       {
        if( (si=s[i]&R)>=k ) 
             if( t[si-k]<N   ) { j=t[sj=s[t[si-k]]&R]; 
             		 		  t[si-k]=j&R|((j&R)>=g?N:0)|j&NN;
            			           t[sj]=i|(j>=N&&j<NNN?N:0); }
        else if( t[si-k]<NN  ) { j=t[ sj = (si-k)   ];
             		 		  t[si-k]=j&R|((j&R)>=g?NNN:0); }
        else if( t[si-k]<NNN ) { 	  t[si-k]=i|NNN; }
        if( s[i]&NN ) g=i;
       }

     for( g=i=0, j=N ; i<n ; i++, C++ ) if( s[i]>=NNN ){ i=s[i]&R; }
     else
       {
        if(  j&NNN  ) g=i;
        si=( s[i]&NN ? s[g=i]&R : (s[j&R]&R)-k);
        j=t[si];
          t[si]=g|( g==i&&(j&NNN) ? NNN : 0 );
       }
    }
 for( i=0 ; i<n ; i++ ) s[t[i]&R]=i;

 printf("n=%d  k=%d  C/n=%d\n",n,k,n?C/n:0);

#if 0
 // Check if sorting is correct
 for( i=0 ; i<n-1 ; i++ )
    {
     si=s[i];
     sj=s[i+1];
     k=n-(si>sj?si:sj);
     for( j=0 ; j<k && b[si+j]==b[sj+j] ; j++ );

     if( si+j==n || j<k && b[si+j]>b[sj+j] )
       { printf("Sorting doesn't work!!!\nsi=%s\nsj=%s\n",b+si,b+sj); exit(1); }
    }
#endif
 // restore original b[] in t[]
 for( g=i=0 ; g<F ; g++ )
   for( ; i<f[g+1] ; i++ ) t[s[i]]=g; // now t[] is equal to original b[]
 
 // Perform BWT based on the sorted string order s[]
 for( i=0 ; i<n ; i++ ) s[i]=s[i]?t[s[i]-1]:F;  s[n]=t[n-1]; n++;
 // now s[n] is a BWT of initial data
 // printf("BWT: ");for( i=0 ; i<n ; i++ ) printf("%c",s[i]<F?s[i]:'$'); printf("\n");

 // Do distance coding
 
 for( i=0 ; i<F+1 ; i++ ) f[i]=n+i;
 for(j=i=0 ; i<n ; i++ )
    {
     k=i;
     for(c=s[i];i<n&&c==s[i];i++);
     i--;
     s[k]=NN+i;
     if( f[c]>=n ) q[j++]=k;
     s[f[c]]=k;
     f[c]=i;
    }
 t[0]=0;m=j;
 for(j=1;j<m;j++) t[j]=q[j]-q[j-1]-1;
 q[j]=NN;
 for(i=0;i<F+1;i++) s[f[i]]=n+i;
 for(i=0;i<n;i++)
    {
     if( s[i]>=NN ) i=s[i]-NN;
     for(k=0;(q[k]=q[k+1])<s[i];k++); q[k]=s[i];
     t[j++]=s[i]-i-k-1;
    }

 // Result of DC: t[0..j-1], m, n
 printf("DC[%d,%d]: ",m,j,(C+7)/8);for(i=0;i<j;i++) printf("%d ",t[i]);printf("\n");

 // unDC t[]
 q[0]=0;
 for(j=1;j<m;j++) q[j]=q[j-1]+t[j]+1;
 for(j=0;j<m;j++) s[q[j]]=j;
 q[j]=n;q[j+1]=NN;
 for(i=0;i<n;)
    {
     for(c=s[i];i<q[1];i++) s[i]=c;
     d=q[1]+t[j++];
     for(k=0;(q[k]=q[k+1])<=d;k++,d++); q[k]=d;
     if(d>=n) f[c]=d-n-1;
     else     s[d]=c;
    }
 for(i=0;i<n;i++) s[i]=f[s[i]];
 //printf("uDC: ");for(i=0;i<n;i++) printf("%c",s[i]<F?s[i]:'$');printf("\n");

 // unBWT s[n] 
 for( i=0 ; i<F+1 ; i++ ) f[i]=0;
 for( i=0 ; i<n   ; i++ ) f[s[i]+1]++;
 for( i=0 ; i<F   ; i++ ) f[i+1]+=f[i];
 for( i=0 ; i<n   ; i++ ) t[f[s[i]]++]=i; j=t[t[n-1]];
 for( i=0 ; i<n ; i++, j=t[j] ) printf("%c",s[j]<F?s[j]:'$'); printf("\n");
 }
}
